import sys,os,pyfits
import shutil,time
import math
from math import pi,sqrt,atan2,cos,sin
import numarray
import numarray.nd_image 
import numarray.convolve as conv
import scipy
import ds9



""" A class to compute the central concentration and asymmetry
paramenters as described in Abraham etal 1994,96, from small
postage stamps and the detection image from the SExtract catalog

Felipe Menanteau, JHU October 2003
Modified to use Kron Radius if defined, JHU. July 2004
Modified to use Petrosian Radius, JHU, May 2005.

"""

class concentration:

    """ Computes the Central Concentration in a galaxy. It can use 3
    types of radius for compute it. This is controled by the radius
    type, Rtype parameter and the geometry param. Rtype can be either
    'Kron','Petro' or None. The geometry param sets the initial shape
    of the ellipse where to include the galaxy's light.

    The geometry can be:
    
    - A tuple with the SExtractor KRON ellipse definition, such as
      geometry = (A_IMAGE,B_IMAGE,THETA). Note that A_IMAGE and
      B_IMAGE are the SExtractor original parameters *multiplied* by
      the KRON_RADIUS param.

    - Or it can be undefined, geometry=None, in that case, it will use
      the pixels within the segmentation mask to compute the centroid,
      and the ellipse shape (a,b,theta) based on the 1st and 2nd
      moments.

    The Rtype can be either  'Kron', 'Petro' or None.
    
    - Kron uses the KRON radius from the SExtractor catalog. So, it
      will use the *exactly* the values of the geometry tuple to
      defined the ellipse mask. In order to use Rtype='Kron' geometry
      must be defined.
    
    - Petro computes the Petrosian radius for eta=0.2, based on the
      geometry tuple in use. It *should* use the SExtractor shape, but
      if geometry=None it would run solo. Not recomended.

    The same properties are passed into the asymmetry class
    """

    def __init__(self,id,fluxfile,segmfile,Rtype='Kron',geometry=None,
                 write=None,verb=None,plot=None):
        
        self.fluxfile = fluxfile
        self.segmfile = segmfile
        self.id       = id
        self.write    = write
        self.outdir   = os.path.dirname(fluxfile)
        self.geometry = geometry
        self.Rtype    = Rtype
        self.verb     = verb
        self.plot     = plot
        self.msg      = ''

        (self.basename,ext) = os.path.splitext(self.fluxfile)    

        # Read in the segmentation and flux images
        read_flux(self)    # Returns self.flux
        read_segment(self) # Returns self.mask, self.seg,
        
        # Area pix and shape
        self.nx   = self.mask.shape[0]
        self.ny   = self.mask.shape[1]

        #
        # If geometry, use its shape,
        # otherwise compute shape from moments
        # 
        if self.geometry:

            if verb: print "# Using input geometry"
            self.A_IMAGE  = self.geometry[0]
            self.B_IMAGE  = self.geometry[1]
            self.THETA    = self.geometry[2]*pi/180.0
            
            # only get the centroid
            self.box = 12
            centroid(self) # Returns xx,yy,xy,a,b,phi

            # Assign geometry as default shape -- for now at least
            self.a   = self.A_IMAGE
            self.b   = self.B_IMAGE
            self.phi = self.THETA
            
        else: # get the moments and the ellipse shape

            # Re-scale a/b to effective area on mask
            self.area = len(self.mask[numarray.where(self.mask == 1)])
            self.box  = max(15,3.0*sqrt(self.area/3.1415))
            self.box  = min(self.box,self.nx*0.95)
            
            moments(self) # Returns xo,yo,xx,yy,xy,a,b,phi
            normalize_ellipse(self) # modifies self.a and self.b

        # 
        # Decide which radius type
        #
        if self.verb: print "# Using Rtype: %s" % Rtype
        if (Rtype == 'Kron' or Rtype == 'kron'):
            
            if not self.geometry:
                print "\n\tERROR: Must provide geometry with %s" % Rtype
                
        elif (Rtype == 'Petro' or Rtype == 'petro'): 

            if self.verb: print "# Computing R_petro"
            r_petro(self,self.flux,plot=self.plot) # Get A_PETRO, B_PETRO, T_PETRO
            
            self.a   = self.A_PETRO
            self.b   = self.B_PETRO
            self.phi = self.T_PETRO
        #else:
        #    print "# WARNING: using segmentation-based area "
            
        # Tuple with the ellipse shape
        shape = (self.xo,self.yo,self.a,self.b,self.phi)
        
        # Make the total and inner elliptical masks
        (self.ell_total) = ellmask(self,shape,1.0,"_ELLTOTAL")
        (self.ell_inner) = ellmask(self,shape,0.3,"_ELLINNER")
        self.ellmask     = self.ell_total
        self.area        = len(self.mask[numarray.where(self.ellmask == 1)])
        self.ellshape    = shape
        
        # And now compute the central concentration
        self.flux_total = self.flux*self.ell_total
        self.flux_inner = self.flux*self.ell_inner
        self.C = self.flux_inner.sum() / self.flux_total.sum()
        
    def showds9(self,nf=4,verb=None):

        ''' Small fuction to plot sub-products using ds9 '''
        
        # Make sure we have ds9
        if not ds9.inpath('ds9',verb=verb):
            print "# WARNING: ds9 not in path"
            return
        
        ds9.prepare(nf,tilemode="grid",verb=verb)
        ds9.put(self.fluxfile,fr=1)
        ds9.ellipse(self.ellshape,color='yellow',units='rad',text=self.Rtype)

        # Draw input geometry -- if available
        if self.geometry:
            geo_shape = (self.xo,self.yo,self.geometry[0],self.geometry[1],self.geometry[2])
            ds9.ellipse(geo_shape,color='blue',text="input",options="background")
            
        # Inner ellipse
        ds9.put(self.basename+"_ELLINNER.fits",fr=2)
        ds9.ellipse(self.ellshape,color='yellow',units='rad')
        
        # Total Ellipse
        ds9.put(self.basename+"_ELLTOTAL.fits",fr=3)
        ds9.ellipse(self.ellshape,color='yellow',units='rad')
        
        ds9.put(self.segmfile,fr=4)
        return
        

class asymmetry(concentration):

    """ Computes the Asymmetry parameter for a galaxy. It bases its
    measurements on the geometry and radius computed by
    concetration. So, asymmetry first call concentration in a dry-run
    to obatin these paramenters."""

    def __init__(self,id,fluxfile,segmfile,Rtype='Kron',
                 geometry=None,write=None,skyrms=None,smooth="no",verb=None,plot=None):

        # Here we only initialize concentration
        concentration.__init__(self,id,fluxfile,segmfile,Rtype,geometry,write=None,plot=plot)

        self.id     = id
        self.write  = write
        self.smooth = smooth
        self.sigma  = 1.0 
        self.skyfactor = 1.0

        # Copy the flux file before smoothing it
        # Useful for keeping the real flux image for later visual checks
        self.bkfile = os.path.join(self.outdir,str(id) + "_flux.fits")
        self.bkfile = self.basename + "_flux.fits"
        shutil.copy(self.fluxfile,self.bkfile);
        
        # Smooth images with gaussian kernel of 1pix width
        if (self.smooth == "yes"):
            gauss(self)
            read_flux(self) # Returns self.flux
            centroid(self)  # Returns xx,yy,xy,a,b,phi
            self.skyfactor = 2.71828 # actually == e, not sure why

        # Compute the sky rms to correct asymmetry if not supplied
        if (skyrms):
            self.skyrms = skyrms/self.skyfactor
        else:
            sky_rms(self)

        # Asymmetry correction factor
        self.corr = sqrt(2)*self.area*self.skyrms

        # Rotate image and masks by 180 degress -- using numarray
        # based function We need a rotated mask, ellipse mask is not
        # really symmetric because of of masked out objects that migh
        # fall in the ellmask
        self.rota      = shrotate(self.flux, 180 ,self.xo, self.yo,
                                  reshape=False, mode='constant',order=1)
        
        self.r_ellmask = shrotate(self.ellmask, 180 ,self.xo, self.yo,
                                  reshape=False, mode='constant',order=1)
        
        # The residual sky pixel values
        sky_pix(self) # Returns self.skypix
        

        # Substract rotated and original
        self.asymm = numarray.abs(self.flux - self.rota)*self.ellmask*self.r_ellmask
        self.total = self.flux * self.ellmask

        # corrected and uncorrected assymetry
        self.Araw = 0.5 *  self.asymm.sum() / self.total.sum()
        self.A    = 0.5 * (self.asymm.sum() - self.corr) / self.total.sum()

        # Write the asymmetry and rotation image into fits
        if (self.write == 'yes'):
            self.asymmfile = self.basename + "_ASYMM.fits"
            self.rotafile  = self.basename + "_ROTA.fits"
            self.ellifile  = self.basename + "_AELLIP.fits"
            self.rllifile  = self.basename + "_RELLIP.fits"
            
            wfits(self.asymm, self.asymmfile, hdr=self.header)
            wfits(self.rota , self.rotafile , hdr=self.header)
            wfits(self.ellmask , self.ellifile , hdr=self.header)
            wfits(self.r_ellmask,self.rllifile , hdr=self.header)


    def showds9(self,nf=4,tilemode='grid',verb=None):
        
        ''' Small fuction to plot sub-products using ds9 '''
        
        xo = self.xo
        yo = self.yo

        # Make sure we have ds9
        if not ds9.inpath('ds9',verb=verb):
            print "# WARNING: ds9 not in path"
            return
        
        ds9.prepare(nf,tilemode="grid",verb=verb)
        ds9.put(self.fluxfile,fr=1)
        ds9.ellipse(self.ellshape,color='yellow',units='rad',text=self.Rtype)
        ds9.cross(xo,yo,color='blue')

        # Draw input geometry -- if available
        if self.geometry:
            geo_shape = (self.xo,self.yo,self.geometry[0],self.geometry[1],self.geometry[2])
            ds9.ellipse(geo_shape,color='blue',text="input",options="background")
            
        # Total Ellipse
        ds9.put(self.basename+"_ELLTOTAL.fits",fr=2)
        ds9.ellipse(self.ellshape,color='yellow',units='rad')
        ds9.cross(xo,yo,color='blue')

        # Asymmetry image
        ds9.put(self.basename + "_ASYMM.fits",fr=3)
        ds9.ellipse(self.ellshape,color='yellow',units='rad')
        ds9.cross(xo,yo,color='blue')

        #ds9.put(self.segmfile,fr=4)
        
        rellipse = self.basename + "_ROTAELL.fits"
        wfits(self.r_ellmask,rellipse,hdr=self.header)
        ds9.put(rellipse,fr=4)
        ds9.ellipse(self.ellshape,color='yellow',units='rad')
        ds9.cross(xo,yo,color='blue')
        
        return




class detection_catalog:

    """Reads in the detection catalog and stores the information"""

    def __init__(self,catalog,verb=None):

        self.catalog = catalog
        if verb:
            print >>sys.stderr, "\r Reading:",catalog
            
        self._read_detection()
        return

    def _read_detection(self):

        self.THETA = {}
        self.X_IMAGE = {}
        self.Y_IMAGE = {}
        self.A_IMAGE = {}
        self.B_IMAGE = {}
        self.KRON_RADIUS = {}
        self.RADIUS = {}
        self.SIZE = {}
        self.BACK = {}
        self.THRESH = {}
        self.FLAG = {}
        self.RMS = {}

        for line in open(self.catalog).readlines():

            vals = line.split()
            
            if vals[0][0] == "#":
                continue
            
            ID = str(vals[0])
            self.X_IMAGE[ID]     = float(vals[1])
            self.Y_IMAGE[ID]     = float(vals[2])
            self.KRON_RADIUS[ID] = float(vals[23])
            self.A_IMAGE[ID]     = float(vals[38])*self.KRON_RADIUS[ID]
            self.B_IMAGE[ID]     = float(vals[39])*self.KRON_RADIUS[ID]
            self.THETA[ID]       = float(vals[42])
            self.RADIUS[ID]      = 1.3*(self.A_IMAGE[ID])
            self.SIZE[ID]        = int(2*self.RADIUS[ID])
            self.BACK[ID]        = float(vals[24])
            self.THRESH[ID]      = float(vals[25])
            self.FLAG[ID]        = int(vals[67])
            self.RMS[ID]         = self.THRESH[ID]/1.5
            
        return


class bpz_catalog:

    """Reads in the photometric redshift catalog and stores the information"""

    def __init__(self,catalog,verb=None):

        self.catalog = catalog
        
        if verb:
            print >>sys.stderr, "\r Reading:",catalog

        self._read_bpz()
        return

    def _read_bpz(self):

        self.Z_B     = {}
        self.Z_B_MIN = {}
        self.Z_B_MAX = {}
        self.T_B     = {}
        self.ODDS    = {}
        self.Z_ML    = {}
        self.T_ML    = {}
        self.CHI2    = {}

        for line in open(self.catalog).readlines():

            vals = line.split()
            
            if vals[0][0] == "#":
                continue

            ID           = str(vals[0])
            self.Z_B[ID]     = float(vals[1])
            self.Z_B_MIN[ID] = float(vals[2])
            self.Z_B_MAX[ID] = float(vals[3])
            self.T_B[ID]     = float(vals[4])
            self.ODDS[ID]    = float(vals[5])
            self.Z_ML[ID]    = float(vals[6])
            self.T_ML[ID]    = float(vals[7])
            self.CHI2[ID]    = float(vals[8])
            
        return

def rotate(self):

    # Rotate image using iraf's rotate, good task, but it requires iraf :(
    from pyraf import iraf
    from pyraf.iraf import images,imgeom
    # Rotate the image with IRAF's rotate
    iraf.imgeom.rotate.xin      = self.xo
    iraf.imgeom.rotate.yin      = self.yo
    iraf.imgeom.rotate.xout     = self.xo
    iraf.imgeom.rotate.yout     = self.yo
    iraf.imgeom.rotate.verbose  = "no"
    iraf.imgeom.rotate.boundary     = "constant"
    iraf.imgeom.rotate.interpolant  = "spline3"
    #iraf.imgeom.rotate.interpolant = "nearest"
    iraf.imgeom.rotate(self.fluxfile,self.rotafile,rotation=180.0)
    
    # Read rotated image in array
    rr = pyfits.open(self.rotafile,"readonly")
    self.rota = rr[0].data
    rr.close()
    return


# Convolve with gaussian using iraf
def gauss(self):
    from pyraf import iraf
    from pyraf.iraf import images
    iraf.images.imfilter.gauss(self.fluxfile,self.fluxfile,self.sigma)
    return


# Read into array the fits file with flux
def read_flux(self):
    ff = pyfits.open(self.fluxfile,"readonly")
    self.flux   = ff[0].data
    self.header = ff[0].header.copy()
    ff.close()
    return self.flux


# Read the segmentation fits file into array
# and make the appropiate mask. If geometry, it creates
# and excluding mask, else in including mask.
def read_segment(self):
    
    # print "Reading ",self.id,"from segm file",self.segmfile
    # Open segmentation fits file
    ss     = pyfits.open(self.segmfile,mode="readonly")
    self.segm   = ss[0].data
    #self.header = ss[0].header.copy()
    ss.close()

    # The number of pixels/area in the segmented image
    # that contain the id in question
    self.segm_area = len( self.segm[self.segm == self.id])


    # The *excluding* mask
    # Mask out pixel from OTHER sources, set them zero, else to one
    self.emask = numarray.where(self.segm == 0       ,1,self.segm)
    self.emask = numarray.where(self.emask == self.id,1,self.emask)
    self.emask = numarray.where(self.emask != 1      ,0,self.emask)
    
    # The *including* mask
    # ONLY use the pixel in the SEGM mask set to 1.0
    self.imask = numarray.where((self.segm == self.id),1,0)
    
    
    if self.geometry:
        self.mask = self.emask
    else:
        self.mask = self.imask
        
    # write the mask into a fits file
    if (self.write == 'yes'):
        maskfile = self.basename + "_MASK.fits"
        wfits(self.mask,maskfile,hdr=self.header)

    return self.mask,self.segm #,self.header

def moments(self):

    bx = self.box

    # Here we get the moments of a (segmented) image
    (j,i) = numarray.indices(self.flux.shape) + 1.0 # add + 1

    # Set the initial centroid
    xc = self.nx/2
    yc = self.ny/2
    # Multiply flux image by mask to compute moments
    # only over the relevant section of code.
    image = self.flux*self.mask
    
    # The centroids first
    # find centroid within a box around (xo,yo)
    #bx = 15 # pixels
    i1 = int(xc - bx/2)
    i2 = int(xc + bx/2)
    j1 = int(yc - bx/2)
    j2 = int(yc + bx/2)

    # Avoid going outside the image
    if i1 < 0 : i1=0
    if j1 < 0 : j1=0
    if i2 > self.nx: i2=self.nx
    if j2 > self.nx: j2=self.nx
    
    x  = i[i1:i2,j1:j2]*image[i1:i2,j1:j2]
    y  = j[i1:i2,j1:j2]*image[i1:i2,j1:j2]

    xx = (i-xc)*(i-xc)*image
    yy = (j-yc)*(j-yc)*image
    xy = (i-xc)*(j-yc)*image

    # This looks more similar to the sextractor's recipe
    # but fail on computing the angles right. We keep the form
    # from Bob's old program instead
    #
    # xx = (i*i)*image
    # yy = (j*j)*image
    # xy = (i*j)*image
    # XX = (xx.sum() / image.sum()) - xc*xc
    # YY = (yy.sum() / image.sum()) - yc*yc
    # XY = (xy.sum() / image.sum()) - xc*yc

    self.xo = x.sum() / image[i1:i2,j1:j2].sum()
    self.yo = y.sum() / image[i1:i2,j1:j2].sum()
    
    self.xx = xx.sum() / image.sum()
    self.yy = yy.sum() / image.sum()
    self.xy = xy.sum() / image.sum()
    
    # Transform the image moments into ellipse parameters
    #
    # This is if we decide to use the trigonometric identity:
    #       tan(2a) = 2tan(a)/(1-tan^2(a))
    # and solve for tan(a) the cuadratic equation, we get
    # two possible solutions for tan(a), the possitive one seems
    # to be the right one.
    # 
    # tan(a) = (yy - xx +/-  sqrt((yy-xx)**2 + 4*xy**2))/(2.*xy)
    #
    # We opt for a simpler one, using half of the angle.
    # Taken from the Sextractor manual:
    # tan(2*phi) = 2 * xy/(xx-yy)
    #

    self.phi = 0.5*atan2(2.*self.xy,(self.xx-self.yy)) # in [rad]
    self.a   = sqrt( 0.5*(self.xx + self.yy + sqrt((self.xx-self.yy)**2 + 4*self.xy**2)))
    self.b   = sqrt( 0.5*(self.xx + self.yy - sqrt((self.xx-self.yy)**2 + 4*self.xy**2)))

    # These are actually half of the values of a/b computed from Bob's
    # iraf code, however they can be re-scaled afterwards using the
    # total area in the segmented image
    
    #return self.xo,self.yo,self.xx,self.yy,self.xy,self.a,self.b,self.phi


def normalize_ellipse(self):
    # Normalize the values of a and b so they represent the same are
    # contained in the segmentation image
    # Get the are of non-zero pixels and the ellipse area

    ell_area = pi*self.a*self.b

    # Store the original values
    self.a_real = self.a
    self.b_real = self.b
    
    # And scale them back
    self.a = self.a * sqrt(self.area/ell_area)
    self.b = self.b * sqrt(self.area/ell_area)
    return self.a,self.b


def ellmask(self,shape,scale,suffix):
    
    xo     = shape[0]
    yo     = shape[1]
    A      = shape[2]
    B      = shape[3]
    THETA  = shape[4]
    
    # Make an elliptical mask from a template image.
    (yi,xi) = numarray.indices(self.mask.shape) + 1.0 # add + 1

    # translate ref center to (xo,yo)    
    x =  (xi-xo)*cos(THETA) + (yi-yo)*sin(THETA)
    y = -(xi-xo)*sin(THETA) + (yi-yo)*cos(THETA)

    # Compute the ellipse and make the mask
    ellipse = x**2/(scale*A)**2 + y**2/(scale*B)**2
    ellmask = numarray.where(ellipse <= 1, 1,0)
    ellmask = ellmask*self.emask # The excluding mask!

    if (self.write == 'yes'):
        ellfile = self.basename + suffix + ".fits"
        wfits(ellmask,ellfile,hdr=self.header)

    return ellmask



# To write in a simple file
def wfits(array,filename,hdr=None):

    """ Writes a simple fits file"""
    
    if os.path.splitext(filename)[-1] <> ".fits":
        filename = filename + ".fits"
        
    # Writes fits files of a given array  
    newfits    = pyfits.HDUList() 
    hdu        = pyfits.PrimaryHDU()
    hdu.data   = array
    if hdr:
         hdu.header = hdr
    newfits.append(hdu)
    
    # Remove old version of file before
    if os.path.isfile(filename):
        os.remove(filename)
    newfits.writeto(filename)
    newfits.close
    return 


# def writefits(self,array,filename):
#     # Writes fits files of a given array 
#     newfits    = pyfits.HDUList()
#     hdu        = pyfits.PrimaryHDU()
#     hdu.header = self.header
#     hdu.header.update('CREATED',time.ctime(),comment="by AC module (F. Menanteau)")

#     hdu.data   = array
#     newfits.append(hdu)

#     # Remove old version of file before
#     if os.path.isfile(filename):
#         os.remove(filename)
        
#     newfits.writeto(filename)
#     return


def sky_rms(self):

    sky = self.flux[numarray.where(self.segm == 0)]
    var = ((sky - sky.mean())**2/len(sky)).sum()
    self.skyrms = sqrt(var)
    return 


def sky_pix(self):

    resima  = numarray.abs(self.flux - self.rota)
    resout  = resima[numarray.where(self.segm == 0)]
    self.skypix = resout.sum()/len(resout)
    return


# And some more generic functions...

def rvals(shape,center=None):
    # Generates radial values from indices

    (yi,xi) = numarray.indices(shape)

    # Get the centre if undefined
    if (center == None):
        xo = int(shape[0]/2)
        yo = int(shape[1]/2)
    else:
        xo = center[0]
        yo = center[1]
        
    # And we make the radial array
    radial = numarray.sqrt( (xi-xo)**2 + (yi-yo)**2)
    return radial


def gauss_kernel(sigma,Nsigma=4):
    # Makes a gaussian kernel of width=sigma and length Nsigma
    # see iraf.gauss procedure
    np = int(2.0*sigma*Nsigma) + 1
    r  = rvals(shape=(np,np))
    kernel = numarray.exp(-0.5*(r/sigma)**2)

    kernel = kernel/kernel.sum()
    return kernel
    


def centroid(self):

    bx = self.box

    (j,i) = numarray.indices(self.flux.shape) + 1.0 # add + 1

    # Set the initial centroid
    xc = self.nx/2
    yc = self.ny/2
    # Multiply flux image by mask to compute moments
    # only over the relevant section of code.
    mask = numarray.where((self.segm == self.id),1,0)
    image = self.flux*mask
    
    # The centroids 
    # find centroid within a box around (xc,yc)
    i1 = int(xc - bx/2)
    i2 = int(xc + bx/2)
    j1 = int(yc - bx/2)
    j2 = int(yc + bx/2)
    
    x  = i[i1:i2,j1:j2]*image[i1:i2,j1:j2]
    y  = j[i1:i2,j1:j2]*image[i1:i2,j1:j2]

    
    self.xo = x.sum() / image[i1:i2,j1:j2].sum()
    self.yo = y.sum() / image[i1:i2,j1:j2].sum()
    
def r_petro(self,image,dr=1.0, nu_0=0.2, plot=None):

    """Routine to get the sum in N rings between a1 and a2 with a
    separation of da, meassured in pixels. It returns the value in
    ring I(r) and the mean values upt to that radius <I(r)>."""


    if (dr < 1.0):
        print "Error: dr must be >= 1.0"
        sys.exit()

    # We have to Decide on the geometry that we are using first
    # to se how we calculate the nu(r) = I(r)/<I(r)>
    
    if self.geometry: # We use the SExtractor ones for rings
        A     = self.A_IMAGE
        B     = self.B_IMAGE
        THETA = self.THETA
        k     = 1.0   # arbitrary scaling factor

    else: # We use the centroids ones for rings
        A     = self.a
        B     = self.b
        THETA = self.phi
        k     = 4.0   # arbitrary scaling factor
        if (A*k > self.nx/2.0):
            k = (self.nx/2.0)/A
            print "# Warning: R_petro, resizing upper k to ", k*A

    # The rest of the input parameters. We put them into private variables
    # to make it easy to migrate to a separate/standalone function in the future
    xo    = self.xo
    yo    = self.yo
    mask  = self.emask # We use the excluding mask

    # Make sure that image and mask have the same shape
    if (image.shape != mask.shape):
        print "Error: mask and image must have same dimensions"
        sys.exit()
    
    # The limits where to measure, we measure based over the
    # semi-major axis all the time.
    a1 = 0.0  # min semi-major
    a2 = A*k  # max semi-major
    
    # Make an elliptical mask from a template image.
    (yi,xi) = numarray.indices(mask.shape) + 1.0 # add + 1
    
    # translate ref center to (xo,yo)    
    x =  (xi-xo)*cos(THETA) + (yi-yo)*sin(THETA)
    y = -(xi-xo)*sin(THETA) + (yi-yo)*cos(THETA)

    # Generate the array of semi-major and semi-minor lengths
    da = dr*A/B
    ax = numarray.arange(a1,a2+da,da) 
    bx = ax * B/A

    # Number of rings/steps
    nr       = len(ax) - 1
    ave_ring = numarray.zeros(nr,'Float32')
    ave_cumu = numarray.zeros(nr,'Float32')
    std_ring = numarray.zeros(nr,'Float32')
    std_cumu = numarray.zeros(nr,'Float32')
    ri       = numarray.zeros(nr,'Float32')

    # the loop...  we make it for each ring
    for i in range(nr):

        # The array with the semi-mayor positions
        ri[i]     = ax[i+1]

        # The ellipse that defines the outter part of ring and total
        # ellipse i.e. the Total ellipse up to r[i]
        ell_upp   = x**2/ax[i+1]**2 + y**2/bx[i+1]**2
        cumu_mask = numarray.where(ell_upp <= 1, 1, 0)

        # The ellipse that defines the inner part of ring and total
        # ellipse for i=0 the ring is the same as the total
        # ellipse. So to avoid dividing by zero, we check first before
        # making the mask i.e the Ring ellipse at ri[i]
        if ( i == 0):
            ell_low   = ell_upp
            ring_mask = cumu_mask
        else:
            ell_low = x**2/ax[i]**2   + y**2/bx[i]**2
            ring_mask = numarray.where(ell_low <= 1, 0, cumu_mask)

        # Times an *excluding* mask -- to avoid other sources' contribution.
        cumu_mask = cumu_mask*mask
        ring_mask = ring_mask*mask

        # The mean 
        ave_cumu[i] = image[cumu_mask==1].mean()
        ave_ring[i] = image[ring_mask==1].mean()

        # The stddev
        std_cumu[i] = image[cumu_mask==1].stddev()
        std_ring[i] = image[ring_mask==1].stddev()

        nring = len(image[ring_mask==1])
        ncumu = len(image[cumu_mask==1])
        ratio = ave_ring[i] / ave_cumu[i]
    
    # Now let's get the r_petro,
    # we use Rp = 1.5*rp
    nu = ave_ring/ave_cumu # I(r)/<I(r)>
    f  = nu - nu_0

    # Make the spline interporation representation
    tck = scipy.interpolate.splrep(ri,f,s=0)

    # We get the root of the spline
    rts = scipy.interpolate.sproot(tck)

    # Check if any of the solucions makes sense
    if (len(rts) < 1):

        # try a minimun
        f_n = f - f.min()
        tck = scipy.interpolate.splrep(ri,f_n,s=0)
        rts = scipy.interpolate.sproot(tck)
        if (len(rts) < 1):
            msg = "# Warning: "+ str(self.id) + " Could not find a solution for r_p, falling back to A:" + A
            print msg
            Rp = A
            rp = A/1.5 # fail safe
        else:
            msg = "# Warning: "+ str(self.id) + " has noisy solution for Rpetro"
            print msg
            rp = rts[0] 
            Rp = 1.5*rp 

        self.msg = msg +"\n"

    else:
        rp = rts[0] # This first solution 
        Rp = 1.5*rp # The SDSS defintion with k=1.5


    # Just to make them visible outside
    self.nu = nu  # I(r)/<I(r)>
    self.ri = ri  # array with r_i

    self.rp      = rp
    self.A_PETRO = Rp
    self.B_PETRO = Rp*B/A
    self.T_PETRO = THETA

    # Plot with ppgplot the solution
    if plot:
        plot_petro(self)
    
    return


def plot_petro(self,nu_0=0.2):

    """ Plots the solution to the eta(r) = 0.2 to obtain the Petrosian
    radius using ppgplot"""
    
    import PYgplot as pygplot

    # Make them locals
    r  = self.ri
    rp = self.rp
    nu = self.nu

    # use Kron if geometry
    if self.geometry:
        Kr = self.A_IMAGE
        rm = numarray.array([r[-1],1.5*rp,Kr])
    else:
        Kr = None
        rm = numarray.array([r[-1],1.5*rp])

    # Plot limits
    r1 = 0.0
    r2 = 1.05*rm.max()
    dr = (r2-r1)/40.

    x = numarray.array([rp,rp])
    y = numarray.array([-1,1])

    u = numarray.array([r1,r2])
    v = numarray.array([nu_0,nu_0])

    xp = 1.5*x

    plot = pygplot.Plot(device="/xs", title='',
                        width=1,
                        linewidth=2,
                        fsize=1.0,
                        yrange=[-0.01,1.02],
                        xrange=[r1,r2],
                        xlabel="r [arcsec]",
                        ylabel='\\gy\\fi(r) = I(r)/\(2227)I(r)\(2228)')

    # The u(r) profile
    plot.point(r,nu)
    plot.line(r,nu)
    # The solution rp where u(r) = 0.2
    plot.line(x,y,color='red',linestyle=4)
    plot.label(rp-dr,     0.95, 'r\\dp\\u',       just=1.0,size=1.0,angle=90.0)

    # Rp = 1.5rp
    plot.line(xp,y,color='red',linestyle=4)
    plot.label(1.5*rp-dr, 0.95, 'R\\dp\\u=1.5r\\dp\\u', just=1.0,size=1.0,angle=90.0)

    plot.line(u,v,color='yellow',linestyle=3)
    plot.label(dr, 0.16, "\\gy\\fi(r)\\fr=0.2", just=0.0,size=1.0,)

    # Kron radius if available
    if Kr:
        kx = numarray.array([Kr,Kr])
        plot.line(kx,y,color='blue',linestyle=4)
        plot.label(Kr-dr, 0.1, 'Kron', just=1.0,size=1.0,angle=90.0)

    plot.label(dr,0.05,str(self.id),just=0.0)
    
    plot.plot()
    plot.close()
    return


def shrotate(input, angle, xin, yin, xout = None, yout = None,
             axes = (-1, -2), reshape = False, output_type = None,
             output = None, order = 3, mode = 'constant', cval = 0.0,
             prefilter = True):


    import numarray
    import numarray.nd_image
    import math

    """Rotate and shift an array around a center.

    Adapted from rotate in numarray, to include a rotation center
    defined by (xin,yin) and and optional output center (xout,yout)
    F Menanteau, JHU. May 2005.

    The array is rotated in the plane definde by the two axes given by
    the axes parameter using spline interpolation of the requested
    order. The angle is given in degrees. Points outside the
    boundaries of the input are filled according to the given
    mode. The output type can optionally be given. If not given it is
    equal to the input type. If reshape is true, the output shape is
    adapted so that the input array is contained completely in the
    output. Optionally an output array can be provided that must match
    the requested output shape and type. The parameter prefilter
    determines if the input is pre-filtered before interpolation, if
    False it is assumed that the input is already filtered.
    
    """
           
    input = numarray.asarray(input)

    angle = numarray.pi / 180 * angle

    if axes[0] < axes[1]:
        a1 = axes[0]
        a2 = axes[1]
    else:
        a1 = axes[1]
        a2 = axes[0]
        
    oshape = list(input.shape)
    ix = input.shape[a1]
    iy = input.shape[a2]
    if reshape:
        # Fix, we now take abs value to avoid crash, when giving
        # negative integer values
        ox = abs(ix * math.cos(angle) + iy * math.sin(angle) + 0.5)
        oy = abs(iy * math.cos(angle) + ix * math.sin(angle) + 0.5)
        ox = int(ox)
        oy = int(oy)
        oshape[a1] = ox
        oshape[a2] = oy
    else:
        ox = oshape[a1]
        oy = oshape[a2]

    m11 = math.cos(angle)
    m12 = math.sin(angle)
    m21 = -math.sin(angle)
    m22 = math.cos(angle)
    matrix = numarray.identity(input.rank, type = numarray.Float64)
    matrix[a1, a1] = m11
    matrix[a1, a2] = m12
    matrix[a2, a1] = m21
    matrix[a2, a2] = m22

    # Fix the indices to start from 1 rather than zero
    xin = xin - 1
    yin = yin - 1

    # Use inputs if output centers not defined
    if xout and yout:
        xout = xout - 1
        yout = yout - 1
    else:
        xout = xin
        yout = yin
     
    offset = numarray.zeros((input.rank,), type = numarray.Float64)
    # offset is just xo,yo
    offset[a1] = yin
    offset[a2] = xin
    # Multiply it by the rotation matrix
    offset = numarray.matrixmultiply(matrix, offset)

    # New center 
    tmp = numarray.zeros((input.rank,), type = numarray.Float64)
    tmp[a1] = yout
    tmp[a2] = xout
    offset = tmp - offset
    
    return numarray.nd_image.affine_transform(input, matrix, offset, oshape, output_type,
                                              output, order, mode, cval, prefilter)
